import logging
from .definitions import *


class Register:
	'''
	Basic class to handle device registers.
	
	A register is a combination of:
	- characteristic
	- data converter for READ
	- data converter for WRITE
	'''
	logger = logging.getLogger(Constants.LOGGER_IO_NAME)


	def __init__(self, name, address, translator_from, translator_to):
		self.name = name
		self.address = address
		self.translator_from = translator_from
		self.translator_to = translator_to


	def from_device(self, value, silent=False):
		'''
		Translates a device byte value to an usable value
		'''
		if not silent:
			self.logger.debug('translating value %s from device register %s', value, self.name)
		v = self.translator_from(value)
		if not silent:
			self.logger.debug('translated value from device register %s to %s', self.name, v)
		return v


	def to_device(self, value, silent=False):
		'''
		Translates the local usable value to device byte value
		'''
		if not silent:
			self.logger.debug('translating value %s to device register %s bytes', value, self.name)
		v = self.translator_to(value)
		if not silent:
			self.logger.debug('translated value to device register %s bytes %s', self.name, v)
		return v


import os
import logging
import pickle
import appdirs

from .definitions import *


class FsUtils:
	'''
	Util class to handle save and restore of application data
	'''
	
	logger = logging.getLogger(Constants.LOGGER_FS_NAME)
	
	def __init__(self, appname):
		self.logger.debug('initializing filestore utils with appname %s', appname)
		self.appauthor = 'LeloF1Py SDK'
		self.appname = appname

	
	def get_connection_profile_path(self):
		'''
		Build the connection profile path for the savefile
		'''
		parent_path = appdirs.user_data_dir(self.appname, self.appauthor)
		os.makedirs(parent_path, exist_ok=True)
		result = parent_path + '/' + 'last_connection_profile.dmp'
		self.logger.debug('connection_profile_path = %s', result)
		return result


	def save_connection_profile(self, profile, safe=False):
		'''
		Attempts to serialize a connection profile to the savefile
		'''
		if not profile:
			raise ValueError('Profile is required.')

		try:
			path = self.get_connection_profile_path()
			self.logger.debug('saving connection profile %s to %s', profile.address, path)
			with open(path, 'wb') as f:
				pickle.dump(profile, f)
				
		except Exception as e:
			self.logger.exception('error while attempting to save connection profile %s to %s: %s', profile.address, path, e)
			if not safe:
				raise e


	def load_connection_profile(self, safe=False):
		'''
		Attempts to load a connection profile from the savefile
		'''
		try:
			path = self.get_connection_profile_path()
			
			if not os.path.exists(path):
				self.logger.debug('no saved connection profile to load in %s', path)
				return None
			
			self.logger.debug('attempting to load connection profile from %s', path)
	
			with open(path, 'rb') as f:
				loaded_obj = pickle.load(f)
				
			return loaded_obj
		
		except Exception as e:
			self.logger.exception('error while attempting to load connection profile from %s: %s', path, e)
			if not safe:
				raise e



class Constants:
	'''
	Application constants and logger names.
	Exposition of logger names allow easy logging configuration
	'''

	LOGGER_NAME = 'lelo-f1-sdk-client'
	LOGGER_IO_NAME = LOGGER_NAME + '.io'
	LOGGER_CALLBACK_NAME = LOGGER_NAME + '.notification'
	LOGGER_SYNC_NAME = LOGGER_NAME + '.sync'
	LOGGER_SOCKET_SERVER_NAME = LOGGER_NAME + '.socket-server'
	LOGGER_FS_NAME = LOGGER_NAME + '.fs'
	ADVERTISING_DEVICE_NAME = 'F1s'


class Characteristics:
	'''
	Contains characteristics identifiers (UUIDs) for the device.
	'''
	KEY_STATE = '00000a0f-0000-1000-8000-00805f9b34fb'
	MOTOR_CONTROL = '0000fff1-0000-1000-8000-00805f9b34fb'
	MANUFACTURER_NAME = '00002a29-0000-1000-8000-00805f9b34fb'
	MODEL_NUMBER = '00002a24-0000-1000-8000-00805f9b34fb'
	HARDWARE_REVISION = '00002a27-0000-1000-8000-00805f9b34fb'
	FIRMWARE_REVISION = '00002a26-0000-1000-8000-00805f9b34fb'
	SOFTWARE_REVISION = '00002a28-0000-1000-8000-00805f9b34fb'
	MAC_ADDRESS = '00000a06-0000-1000-8000-00805f9b34fb'
	SERIAL_NUMBER = '00000a05-0000-1000-8000-00805f9b34fb'
	BATTERY_LEVEL = '00002a19-0000-1000-8000-00805f9b34fb'
	MOTOR_WORK_ON_TOUCH = '00000aa5-0000-1000-8000-00805f9b34fb'
	VIBRATOR_SETTING = '00000a0d-0000-1000-8000-00805f9b34fb'
	WAKE_UP	= '00000aa1-0000-1000-8000-00805f9b34fb'
	HALL = '00000aa3-0000-1000-8000-00805f9b34fb'
	LENGTH = '00000a0b-0000-1000-8000-00805f9b34fb'
	ACCELEROMETER = '00000a0c-0000-1000-8000-00805f9b34fb'
	PRESSURE = '00000a0a-0000-1000-8000-00805f9b34fb'
	BUTTON = '00000aa4-0000-1000-8000-00805f9b34fb'
	USER_RECORD	= '00000a04-0000-1000-8000-00805f9b34fb'
	CHIP_ID = '00000a07-0000-1000-8000-00805f9b34fb'
	
	# unreadable (err. 2)
	BATTERY_VOLTAGE = '00000a00-0000-1000-8000-00805f9b34fb'
	OTA	= '00000a08-0000-1000-8000-00805f9b34fb'
	
	# reads [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
	ACTIVATE = '00000a0e-0000-1000-8000-00805f9b34fb'

	# reads [0]
	ACCELEROMETER_CONTROL = '00000aa0-0000-1000-8000-00805f9b34fb'

	# reads [0]
	HALL_CONTROL = '00000aa2-0000-1000-8000-00805f9b34fb'

	# ServiceName: GenericAccess
	# CharacteristicName: DeviceName
	# reads [70, 49, 115] "F1s"
	GENERIC_ACCESS_DEVICE_NAME = '00002a00-0000-1000-8000-00805f9b34fb'
	
	# ServiceName: GenericAccess
	# CharacteristicName: Appearance
	# reads [0, 0]
	GENERIC_ACCESS_APPEARANCE = '00002a01-0000-1000-8000-00805f9b34fb'
	
	# ServiceName: GenericAccess
	# CharacteristicName: PeripheralPreferredConnectionParameters
	# reads [80, 0, 160, 0, 0, 0, 232, 3]
	GENERIC_ACCESS_PERIPHERAL_PREFERRED_CONNECTION_PARAMETERS = '00002a04-0000-1000-8000-00805f9b34fb'
	
	# ServiceName: DeviceInformation
	# CharacteristicName: SystemId
	# reads [238, 91, 69, 0, 0, 227, 100, 196]
	DEVICE_INFORMATION_SYSTEM_ID = '00002a23-0000-1000-8000-00805f9b34fb'
	
	# ServiceName: DeviceInformation
	# CharacteristicName: SerialNumberString
	# reads [83, 101, 114, 105, 97, 108, 32, 78, 117, 109, 98, 101, 114] "Serial Number"
	DEVICE_INFORMATION_SERIAL_NUMBER_STRING = '00002a25-0000-1000-8000-00805f9b34fb'
	
	# ServiceName: DeviceInformation
	# CharacteristicName: Ieee11073_20601RegulatoryCertificationDataList
	# reads [254, 0, 101, 120, 112, 101, 114, 105, 109, 101, 110, 116, 97, 108]
	DEVICE_INFORMATION_IEEE11073 = '00002a2a-0000-1000-8000-00805f9b34fb'
	
	# ServiceName: DeviceInformation
	# CharacteristicName: PnpId
	# reads [1, 13, 0, 0, 0, 16, 1]
	DEVICE_INFORMATION_PNP_ID = '00002a50-0000-1000-8000-00805f9b34fb'


class Services:
	'''
	Contains services identifiers (UUIDs) for the device.
	Unused at the moment.
	'''
	GENERIC_ACCESS_PROFILE = '00001800-0000-1000-8000-00805f9b34fb' 
	GENERIC_ATTRIBUTE_PROFILE = '00001801-0000-1000-8000-00805f9b34fb' 
	DEVICE_INFORMATION = '0000180a-0000-1000-8000-00805f9b34fb'
	VENDOR_SPECIFIC = '0000fff0-0000-1000-8000-00805f9b34fb'
	BATTERY_SERVICE = '0000180f-0000-1000-8000-00805f9b34fb' 


class CruiseControlStatus:
	'''
	Alias for Cruise Control status.
	For internal use only: value is translated to boolean when accessed from client methods.
	Not that values ENABLE_AND_RESET supports write only
	'''
	DISABLED = 0x00
	ENABLED = 0x01
	ENABLE_AND_RESET = 0x02


class WakeUp:
	'''
	Alias for quick Wake-Up status.
	For internal use only: value is translated to boolean when accessed from client methods.
	'''
	DISABLED = 0x00
	ENABLED = 0x01


class Buttons:
	'''
	Alias for buttons status.
	'''
	NONE_PRESSED = 0x03
	CENTRAL = 0x00
	PLUS = 0x01
	MINUS = 0x02


class ConnectionProfile:
	'''
	Holds information on connected device
	'''
	address = None
	uuid = None
	name = None


import logging
from .definitions import *


SYNC_LOGGER = logging.getLogger(Constants.LOGGER_SYNC_NAME)


def synchronized(lock):
	'''
	Synchronization decorator to manage concurrent access to critical resources
	'''
	
	def synchronized_decorator(f):
		def synchronized_wrapper(*args, **kw):
			_resource_id = lock.__hash__()
			SYNC_LOGGER.debug("acquiring lock for resource {}".format(_resource_id))
			lock.acquire()
			SYNC_LOGGER.debug("lock acquired for resource {}".format(_resource_id))
			try:
				return f(*args, **kw)
			finally:
				SYNC_LOGGER.debug("lock released for resource {}".format(_resource_id))
				lock.release()
		return synchronized_wrapper
	return synchronized_decorator


from .register import *
from .definitions import *


def not_supported(value):
	'''
	Shortcut to inform that the selected READ or WRITE operation is not supported on the register.
	'''
	raise ValueError('Operation not supported from register')


class Converters:
	'''
	Standard data converters to be reused in register definitions
	'''
	ENCODING = 'UTF-8'

	BOOLEAN_CONVERTER_FROM = lambda device_value : True if (device_value[0] & 0x01) else False
	BOOLEAN_CONVERTER_TO = lambda local_value : [0x01 if local_value else 0x00]
	NOT_SUPPORTED = lambda value: not_supported(value)
	ASCII = lambda value: value.decode(Converters.ENCODING)
	HEX_STRING = lambda bytes: ':'.join(['%02x'%b for b in bytes]).upper()
	SINGLE_INTEGER = lambda value: value[0]
	SINGLE_16B_INTEGER = lambda value: value[0]*256+value[1]
	INTEGER_ARRAY = lambda value: [int(v) for v in value]
	INTEGER_TUPLE = lambda value: tuple([int(v) for v in value]) 


class Registers:
	'''
	Contains all registers definitions.

	A register is a combination of:
	- characteristic
	- data converter for READ
	- data converter for WRITE
	
	This layer prevents the main client from having to deal with low-level data conversion.
	'''
	
	KEY_STATE = Register(
		'KEY STATE', Characteristics.KEY_STATE,
		Converters.BOOLEAN_CONVERTER_FROM, Converters.NOT_SUPPORTED )

	MOTOR_SPEED = Register(
		'MOTOR SPEED', Characteristics.MOTOR_CONTROL, 
		lambda device_value : tuple([device_value[1], device_value[2]]),
		lambda local_value : bytes([0x01, local_value[0], local_value[1]]) )

	MOTOR_STOP = Register(
		'MOTOR STOP', Characteristics.MOTOR_CONTROL, 
		Converters.NOT_SUPPORTED, lambda local_value : [0x01, 0x00, 0x00] )

	SHUTDOWN = Register(
		'SHUTDOWN', Characteristics.MOTOR_CONTROL, 
		Converters.NOT_SUPPORTED, lambda local_value : [0x01, 0xFA] )

	VERIFY_ACCELEROMETER = Register(
		'VERIFY ACCELEROMETER', Characteristics.MOTOR_CONTROL, 
		Converters.NOT_SUPPORTED, lambda local_value : [0xFF, 0xFF, 0xFF] )

	MOTOR_WORK_ON_TOUCH = Register(
		'MOTOR WORK ON TOUCH', Characteristics.MOTOR_WORK_ON_TOUCH, 
		lambda device_value : True if device_value[0] == CruiseControlStatus.ENABLED else False, 
		lambda local_value : [local_value] )

	VIBRATOR_SETTING = Register(
		'VIBRATOR SETTING', Characteristics.VIBRATOR_SETTING, 
		Converters.INTEGER_TUPLE, Converters.INTEGER_ARRAY )

	WAKE_UP = Register(
		'WAKE UP', Characteristics.WAKE_UP, 
		lambda device_value : True if device_value[0] == WakeUp.ENABLED else False, 
		lambda local_value : [WakeUp.ENABLED if local_value else WakeUp.DISABLED] )

	HALL = Register(
		'HALL', Characteristics.HALL, 
		Converters.SINGLE_16B_INTEGER, Converters.NOT_SUPPORTED )

	LENGTH = Register(
		'LENGTH', Characteristics.LENGTH, 
		Converters.SINGLE_16B_INTEGER, Converters.NOT_SUPPORTED )

	ACCELEROMETER = Register(
		'ACCELEROMETER', Characteristics.ACCELEROMETER, 
		lambda device_value : tuple([
			device_value[0]*256+device_value[1],
			device_value[2]*256+device_value[3],
			device_value[4]*256+device_value[5] ]), 
		Converters.NOT_SUPPORTED )

	PRESSURE_TEMPERATURE = Register(
		'PRESSURE_TEMPERATURE', Characteristics.PRESSURE, 
		lambda device_value : tuple([
			(device_value[0]*256*256+device_value[1]*256+device_value[2]) / 100.0,
			(device_value[4]*256*256*256+device_value[5]*256*256+device_value[6]*256+device_value[7]) / 100.0 ]), 
		Converters.NOT_SUPPORTED )

	BUTTON =  Register(
		'BUTTON', Characteristics.BUTTON, 
		Converters.SINGLE_INTEGER, Converters.NOT_SUPPORTED )

	USER_RECORD =  Register(
		'USER RECORD', Characteristics.USER_RECORD, 
		Converters.SINGLE_16B_INTEGER, Converters.NOT_SUPPORTED )

	USER_RECORD_RESET =  Register(
		'USER RECORD RESET', Characteristics.USER_RECORD,
		Converters.NOT_SUPPORTED, 
		lambda local_value : [0xEE] )

	MANUFACTURER_NAME = Register(
		'MANUFACTURER NAME', Characteristics.MANUFACTURER_NAME, 
		Converters.ASCII, Converters.NOT_SUPPORTED )

	MODEL_NUMBER = Register(
		'MODEL NUMBER', Characteristics.MODEL_NUMBER, 
		Converters.ASCII, Converters.NOT_SUPPORTED )
	
	HARDWARE_REVISION = Register( 
		'HARDWARE REVISION', Characteristics.HARDWARE_REVISION, 
		Converters.ASCII, Converters.NOT_SUPPORTED )

	FIRMWARE_REVISION = Register( 
		'FIRMWARE REVISION', Characteristics.FIRMWARE_REVISION, 
		Converters.ASCII, Converters.NOT_SUPPORTED )

	SOFTWARE_REVISION = Register( 
		'SOFTWARE REVISION', Characteristics.SOFTWARE_REVISION, 
		Converters.ASCII, Converters.NOT_SUPPORTED )

	MAC_ADDRESS = Register( 
		'MAC ADDRESS', Characteristics.MAC_ADDRESS, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	
	SERIAL_NUMBER = Register( 
		'SERIAL NUMBER', Characteristics.SERIAL_NUMBER, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	
	BATTERY_LEVEL = Register( 
		'BATTERY LEVEL', Characteristics.BATTERY_LEVEL, 
		Converters.SINGLE_INTEGER, Converters.NOT_SUPPORTED )
		
	CHIP_ID = Register( 
		'CHIP ID', Characteristics.CHIP_ID, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	
	GENERIC_ACCESS_DEVICE_NAME = Register( 
		'DEVICE NAME', Characteristics.GENERIC_ACCESS_DEVICE_NAME, 
		Converters.ASCII, Converters.NOT_SUPPORTED )
	
	GENERIC_ACCESS_APPEARANCE = Register( 
		'APPEARANCE', Characteristics.GENERIC_ACCESS_APPEARANCE, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	
	GENERIC_ACCESS_PERIPHERAL_PREFERRED_CONNECTION_PARAMETERS = Register( 
		'PPCP', Characteristics.GENERIC_ACCESS_PERIPHERAL_PREFERRED_CONNECTION_PARAMETERS, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	 
	DEVICE_INFORMATION_SYSTEM_ID = Register( 
		'SYSTEM ID', Characteristics.DEVICE_INFORMATION_SYSTEM_ID, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	
	DEVICE_INFORMATION_SERIAL_NUMBER_STRING = Register( 
		'SERIAL NUMBER STRING', Characteristics.DEVICE_INFORMATION_SERIAL_NUMBER_STRING, 
		Converters.ASCII, Converters.NOT_SUPPORTED )
	
	DEVICE_INFORMATION_IEEE11073 = Register( 
		'IEEE11073', Characteristics.DEVICE_INFORMATION_IEEE11073, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	
	DEVICE_INFORMATION_PNP_ID = Register( 
		'PNP ID', Characteristics.DEVICE_INFORMATION_PNP_ID, 
		Converters.HEX_STRING, Converters.NOT_SUPPORTED )
	



import asyncio
import time
import logging
import socket
import json


from .async_client import *


class SocketClient(AsyncClient):
	'''
	Asynchronous LELO F1 SDK SOCKET client.
	Configurable with standard logging.
	'''
	SOCKET_ENCODING = 'utf8'

	def __init__(self, ip_address, port):
		'''
		Instantiate the client. Takes ip_address as string and port as integer
		'''
		self.logger.debug('instantiating LELO F1 SDK SOCKET client')
		self.adapter = None
		
		self.remote_address = ip_address
		self.remote_port = port
		self.socket = None
		

	async def discover(self, timeout=1, address=None):
		raise ValueError('DISCOVER is not supported on socket client')


	async def connect(self):
		'''
		Connects to device with given ip_address and port.
		The 'timeout' parameter specifices maximum time before giving up.
		'''
		if not self.remote_address or not self.remote_port:
			raise ValueError('Address is required. Please provide ip_address and port')
		
		self.logger.info('attempting connection to %s:%d', self.remote_address, self.remote_port)
		
		try:
			self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.socket.connect((self.remote_address, self.remote_port))
			
			self.logger.info('succesfully connected to %s:%d', self.remote_address, self.remote_port)
			self.connected = True
			self.connected_address = self.remote_address + ':' + str(self.remote_port)
		
		except Exception as e:
			self.logger.exception('connection attempt failed: %s', e)
			self.connected = False
			raise e


		# Pings (reading a DeviceInfo register) to verify communication
		self.logger.debug('verifying communication layer')
		await self.ping()

		return self.socket
	
	
	async def disconnect(self):
		'''
		Disconnects from device.
		'''
		if not self.connected:
			self.logger.warning('client is not connected')
			return
		
		await self._disconnect()
	
	
	async def shutdown(self):
		'''
		Shutdown the device and disconnects.
		Automatically cancels pending notifications.
		'''
		await self.assert_authorized()

		self.logger.info('stopping motors')
		await self.stop_motors()
		
		await self._disconnect()


	async def _disconnect(self):
		self.logger.info('disconnecting from device')

		self.socket.close()

		self.connected = False
		self.socket = None
		self.connected_address = None

		self.logger.info('disconnected from device')


	async def is_authorized(self):
		'''
		Checks wether the user has authorized the connection by pressing the central button
		'''
		return self.is_connected() and (await self.get_remote_authorization() == True) and await self.get_key_state(silent=True)


	async def is_connection_authorized(self):
		'''
		Checks wether the user has authorized the connection by pressing the central button
		'''
		return self.is_connected() and (await self.get_remote_authorization() == True)


	async def is_remote_connected(self):
		'''
		Checks wether the user has authorized the connection by pressing the central button
		'''
		return self.is_connected() and await self.get_remote_connected()


	async def get_remote_authorization(self):
		'''
		Checks if remote server has authorized this client connection.
		'''
		return await self.send_command('authorized')


	async def get_remote_connected(self):
		'''
		Checks if remote server has connected to device.
		'''
		return await self.send_command('connected')


	async def request_remote_authorization(self):
		'''
		Prompts the server for authorization.
		'''
		return await self.send_command('authorize')


	async def send_command(self, command, arguments=None):
		'''
		Sends a command to remote socket.
		Handles JSON serialization/deserialization
		'''
		self.assert_connected()

		payload = dict()
		payload['command'] = command
		if arguments is not None:
			payload['arguments'] = arguments
		
		payload_str = json.dumps(payload)
		
		self.logger_io.debug('SEND %s', payload_str)
		
		self.socket.send(payload_str.encode(self.SOCKET_ENCODING))
		raw_response = self.socket.recv(1023).decode(self.SOCKET_ENCODING)
		
		self.logger_io.debug('RECEIVED %s', raw_response)
		
		response = json.loads(raw_response)

		if response and 'status' in response:
			if response['status'] == 'OK':
				return response['data'] if 'data' in response else None
			else:
				msg = response['status'] + ' - ' + (response['message'] if 'message' in response else 'generic error')
				self.logger_io.error('Response KO: %s - %s', raw_response, msg)
				raise ValueError(msg)
		else:
			self.logger_io.error('Malformed response: %s', raw_response)
			raise ValueError('Malformed response')
		

	async def assert_authorized(self):
		'''
		Shortcut method to raise error if the connection has not been authorized by pressing the central button
		'''
		if not await self.is_authorized():
			raise ValueError('Client is not authorized. Remote user should explicitly authorize you.')


	async def set_name(self, name):
		'''
		Sets the client name.
		'''
		return await self.send_command('name', [name])


	async def get_manufacturer_name(self):
		'''
		Reads the manufacturer name.
		Returns an ASCII UTF-8 string
		'''
		return await self.send_command('get_manufacturer_name')


	async def get_model_number(self):
		'''
		Reads the model number.
		Returns an ASCII UTF-8 string
		'''
		return await self.send_command('get_model_number')


	async def get_hardware_revision(self):
		'''
		Reads the device hardware revision.
		Returns an ASCII UTF-8 string
		'''
		return await self.send_command('get_hardware_revision')


	async def get_firmware_revision(self):
		'''
		Reads the device firmware revision.
		Returns an ASCII UTF-8 string
		'''
		return await self.send_command('get_firmware_revision')


	async def get_software_revision(self):
		'''
		Reads the device software revision.
		Returns an ASCII UTF-8 string
		'''
		return await self.send_command('get_software_revision')


	async def get_mac_address(self):
		'''
		Reads the device mac address.
		Returns an ASCII UTF-8 string in the format AA:BB:CC:DD:EE:FF
		'''
		return await self.send_command('get_mac_address')


	async def get_serial_number(self):
		'''
		Reads the device serial number.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		return await self.send_command('get_serial_number')


	async def get_chip_id(self):
		'''
		Reads the device CHIP ID.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		return await self.send_command('get_chip_id')


	async def get_device_name(self):
		'''
		Reads the device name.
		Returns an ASCII UTF-8 string
		'''
		return await self.send_command('get_device_name')


	async def get_system_id(self):
		'''
		Reads the system id.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		return await self.send_command('get_system_id')


	async def get_pnp_id(self):
		'''
		Reads the PNP id.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		return await self.send_command('get_pnp_id')


	async def get_ieee_11073_20601(self):
		'''
		Reads the IEEE 11073 20601 regulation mandatory specification.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		return await self.send_command('get_ieee_11073_20601')


	async def get_battery_level(self):
		'''
		Reads the battery level.
		Returns an integer from 0 to 100
		'''
		return await self.send_command('get_battery_level')


	async def get_use_count(self):
		'''
		Reads the use count.
		Returns an integer.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_use_count')


	async def reset_use_count(self):
		'''
		Resets device use count to zero.
		'''
		# await self.assert_authorized()
		self.logger.debug('resetting use count')
		return await self.send_command('reset_use_count')


	async def get_buttons_status(self):
		'''
		Reads the button status. 
		Returns an int value from the enum Buttons:
			Buttons.NONE_PRESSED = 0x03
			Buttons.CENTRAL = 0x00
			Buttons.PLUS = 0x01
			Buttons.MINUS = 0x02
		'''
		# await self.assert_authorized()
		return await self.send_command('get_buttons_status')


	async def get_temperature_and_pressure(self):
		'''
		Reads the device internal temperature and pressure.
		Returns a touple (temperature, pressure).
		Temperature is a floating point number with 2 decimals in Celsius degrees.
		Pressure is a floating point number with 2 decimals in mbar.
		Eg. (24.25, 987.23)
		'''
		# await self.assert_authorized()
		return await self.send_command('get_temperature_and_pressure')


	async def get_temperature(self):
		'''
		Reads the device internal temperature.
		Returns a floating point number with 2 decimals in Celsius degrees.
		Eg. (24.25, 987.23)
		'''
		# await self.assert_authorized()
		return await self.send_command('get_temperature')


	async def get_pressure(self):
		'''
		Reads the device internal pressure.
		Returns a floating point number with 2 decimals in mbar.
		Eg. 987.23
		'''
		# await self.assert_authorized()
		return await self.send_command('get_pressure')


	async def get_accelerometer(self):
		'''
		Reads the device accelerometer data.
		Returns a touple (x, y, z) where x, y and z are integers.
		Eg. (10, 200, 1000)
		'''
		# await self.assert_authorized()
		return await self.send_command('get_accelerometer')


	async def get_accelerometer_x(self):
		'''
		Reads the device accelerometer data for X axis.
		Returns an integer.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_accelerometer_x')


	async def get_accelerometer_y(self):
		'''
		Reads the device accelerometer data for Y axis.
		Returns an integer.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_accelerometer_y')


	async def get_accelerometer_z(self):
		'''
		Reads the device accelerometer data for Z axis.
		Returns an integer.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_accelerometer_z')


	async def get_depth(self):
		'''
		Reads the insertion depth level.
		Returns an integer from 0 to 8.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_depth')


	async def get_rotation_speed(self):
		'''
		Reads the rotation speed from Hall sensors.
		Returns an integer representing the rotations per second.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_rotation_speed')


	async def get_wake_up(self):
		'''
		Reads the quick wake-up status.
		Returns a boolean.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_wake_up')


	async def enable_wake_up(self):
		'''
		Enables quick wake-up.
		'''
		# await self.assert_authorized()
		self.logger.info('setting wake-up to %s', value)
		return await self.send_command('enable_wake_up')


	async def disable_wake_up(self):
		'''
		Disables quick wake-up.
		'''
		# await self.assert_authorized()
		self.logger.info('setting wake-up to %s', value)
		return await self.send_command('disable_wake_up')


	async def get_vibration_setting(self):
		'''
		Reads the device auto vibration settings
		Returns a touple of 8 integers ranging from 0 to 100.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_vibration_setting')


	async def set_vibration_setting(self, value):
		'''
		Sets the auto vibration settings.
		Takes a collection of 8 integers ranging from 0 to 100.
		'''
		if not value:
			raise ValueError('Value is required')
		if len(value) != 8:
			raise ValueError('A collection of 8 elements is required')
		for v in value:
			if v < 0 or v > 100:
				raise ValueError('Value should be between 0 and 100')
		
		# await self.assert_authorized()
		self.logger.info('setting vibration-setting to %s', value)
		return await self.send_command('set_vibration_setting', [value])


	async def get_cruise_control(self):
		'''
		Reads the cruise control status.
		Returns a boolean.
		'''
		# await self.assert_authorized()
		return await self.send_command('get_cruise_control')


	async def disable_cruise_control(self):
		'''
		Disables cruise control.
		'''
		# await self.assert_authorized()
		return await self.send_command('disable_cruise_control')


	async def enable_cruise_control(self, reset=False):
		'''
		Enables cruise control.
		If reset=True is passed, resets motors speed to default value.
		'''
		# await self.assert_authorized()
		return await self.send_command('enable_cruise_control', [reset])


	async def get_key_state(self, silent=False):
		'''
		Reads the key state status.
		Returns a boolean.
		'''		
		self.assert_connected()
		if not silent:
			self.logger.debug('checking key state')
		return await self.send_command('get_key_state')


	async def stop_motors(self):
		'''
		Stops both motors of the device.
		'''
		# await self.assert_authorized()
		
		self.logger.info('sending motors stop signal')
		return await self.send_command('stop_motors')


	async def verify_accelerometer(self):
		'''
		Puts the device into VERIFY ACCELEROMETER mode.
		'''
		# await self.assert_authorized()
		self.logger.info('enterying accelerometer verification mode')
		return await self.send_command('verify_accelerometer')
	
	
	async def get_motors_speed(self):
		'''
		Reads the two motors speed.
		Returns a touple (X, Y) with X being the main motor speed ranging from 0 to 100
		and Y being the vibrator motor speed ranging from 0 to 100
		'''
		# await self.assert_authorized()
		return await self.send_command('get_motors_speed')


	async def get_main_motor_speed(self):
		'''
		Reads the main motor speed.
		Returns an integer ranging from 0 to 100
		'''
		# await self.assert_authorized()
		return await self.send_command('get_main_motor_speed')


	async def get_vibration_speed(self):
		'''
		Reads the vibrator motor speed.
		Returns an integer ranging from 0 to 100
		'''
		# await self.assert_authorized()
		return await self.send_command('get_vibration_speed')


	async def set_motors_speed(self, value):
		'''
		Set the two motors speed.
		Takes a collection of two elements (X, Y) 
		with X being the main motor speed ranging from 0 to 100
		and Y being the vibrator motor speed ranging from 0 to 100
		'''
		if not value:
			raise ValueError('Value is required')
		if len(value) != 2:
			raise ValueError('A collection of 2 elements is required')
		for v in value:
			if v < 0 or v > 100:
				raise ValueError('Value should be between 0 and 100')
		
		# await self.assert_authorized()
		self.logger.info('setting motor speed to %d and vibration speed to %d', value[0], value[1])
		return await self.send_command('set_motors_speed', [value])


	async def set_main_motor_speed(self, value):
		'''
		Sets the main motor speed.
		Takes an integer ranging from 0 to 100
		'''
		if value is None:
			raise ValueError('Value is required')
		if value < 0 or value > 100:
			raise ValueError('Value should be between 0 and 100')
		
		# await self.assert_authorized()
		self.logger.info('setting main motor speed to %d', value)
		return await self.send_command('set_main_motor_speed', [value])


	async def set_vibration_speed(self, value):
		'''
		Sets the vibrator motor speed.
		Takes an integer ranging from 0 to 100
		'''
		if value is None:
			raise ValueError('Value is required')
		if value < 0 or value > 100:
			raise ValueError('Value should be between 0 and 100')
		
		# await self.assert_authorized()
		self.logger.info('setting vibration speed to %d', value)
		return await self.send_command('set_vibration_speed', [value])


	async def ping(self):
		'''
		Reads from a register to test device communication
		'''
		self.assert_connected()

		self.logger.debug('sending ping to device')
		await self.send_command('ping')
		self.logger.debug('device responded to ping')
	

	async def read(self, register, silent=False): self.not_supported()
	async def write(self, register, value=None): self.not_supported()
	async def notify_key_state(self, user_callback, distinct_until_changed=True): self.not_supported()
	async def notify_buttons(self, user_callback, distinct_until_changed=True): self.not_supported()
	async def notify_rotation_speed(self, user_callback, distinct_until_changed=True): self.not_supported()
	async def notify_depth(self, user_callback, distinct_until_changed=True): self.not_supported()
	async def notify_accelerometer(self, user_callback, distinct_until_changed=True): self.not_supported()
	async def notify_temperature_and_pressure(self, user_callback, distinct_until_changed=True): self.not_supported()
	async def _create_callback_handler(self, register, user_callback, distinct_until_changed=False): self.not_supported()
	def _dispatch_notification(self, register, sender, raw_data, distinct_until_changed=False): self.not_supported()
	async def unregister(self, callback_handler): self.not_supported()
	async def _unregister(self, callback_handler): self.not_supported()
	async def _shutdown_notifications(self): self.not_supported()
	
	def not_supported(self):
		raise ValueError('Not supported on socket client')


import asyncio
import bleak
import time
import logging
import threading

from .definitions import *
from .registers import *
from .callback_handler import *
from .helpers import *

# Synchronization handle for concurrent resource access
SYNC_LOCK = threading.Lock()


class AsyncClient(object):
	'''
	Asynchronous LELO F1 SDK client.
	Configurable with standard logging.
	BLE backend is based on Bleak module.
	'''
	
	logger = logging.getLogger(Constants.LOGGER_NAME)
	logger_io = logging.getLogger(Constants.LOGGER_IO_NAME)
	logger_callback = logging.getLogger(Constants.LOGGER_CALLBACK_NAME)

	adapter = None
	connected = False
	connected_address = None
	bleak_client = None
	user_handlers = dict()
	bleak_handlers = dict()
	callback_data_history = dict()
	
	key_state_check = True


	def __init__(self):
		'''
		Instantiate the client. Takes no parameters.
		'''
		self.logger.debug('instantiating LELO F1 SDK client')
		self.adapter = bleak


	async def discover(self, timeout=1, address=None):
		'''
		Look for F1s devices. Filters by name == 'F1s'
		If an address is provided, it filters by address instead.
		The 'timeout' parameter specifices maximum time before giving up.
		'''
		found = []
		time_start = time.time()
		
		while not len(found) and time.time() - time_start < timeout:
			self.logger.debug('discovering nearby devices')
			all_devices = await self.adapter.discover()
			
			self.logger.debug('found %d nearby devices', len(all_devices))
			if address is not None:
				# filter by address
				for device in all_devices:
					if device.address == address:
						self.logger.info('discovery found device with correct address: %s', device.name)
						found.append(device)
				self.logger.debug('of which %i with correct address', len(found))

			else:
				# filter by name	
				for device in all_devices:
					if device.name == Constants.ADVERTISING_DEVICE_NAME:
						self.logger.info('discovery found device with correct advertising name: %s, address=%s', device.name, device.address)
						found.append(device)
				self.logger.debug('of which %i with correct name', len(found))

		self.logger.debug('found a total of %i devices responding to criteria', len(found))
		self.logger.debug('discovery took %i seconds', int(time.time() - time_start))

		return found


	async def connect(self, address, timeout=1):
		'''
		Connects to device with given MAC address.
		The 'timeout' parameter specifices maximum time before giving up.
		'''
		if not address:
			raise ValueError('Address is required. Please provide MAC address of target device')
		time_start = time.time()
		
		# Using Bleak as backend
		client = bleak.BleakClient(address)
		connected = False
		err = None
		attempt_num = 0
		
		self.logger.info('connecting to device %s', address)

		while not connected and err is None:
			# Reattempts until timeout
			attempt_num = attempt_num + 1
			self.logger.debug('attempting connection to device with address %s', address)
			try:
				await client.connect()
				connected = True
			except Exception as e:
				self.logger.debug("connection attempt #%d failed: %s", attempt_num, e)
				if time.time() - time_start < timeout:
					# Sleep 250ms between attempts
					await asyncio.sleep(0.25)
				else:
					err = e
					self.logger.warning('connection attempt timed out')
		
		if not connected:
			# Raise last exception when giving up
			raise err if err is not None else ValueError('connection failed in an unexpected way')

		self.logger.info('succesfully connected to device %s', address)
		self.connected = True
		self.bleak_client = client
		self.connected_address = address
		
		# Profiles
		self.logger.debug('profiling the device')
		try:
			await self._profile_device()
		except Exception as e:
			self.logger.exception('error profiling device: %s', e)
		
		# Pings (reading a DeviceInfo register) to verify communication
		self.logger.debug('verifying communication layer')
		await self.ping()

		return self.bleak_client
	
	
	async def _profile_device(self):
		'''
		Enumerates characteristics on the device
		'''
		profiling = await self.bleak_client.get_services()
		for k, v in profiling.characteristics.items():
			try:
				read_val = await self.bleak_client.read_gatt_char(k)
				read_val = ':'.join(['%02x'%b for b in read_val]).upper()
			except Exception as e:
				read_val = 'cannot read: ' + str(e)
			self.logger.debug('root characteristic %s -> %s = [%s]', k, v, read_val)

		for k, v in profiling.services.items():
			self.logger.debug('service %s -> %s', k, v.description)
			for v2 in v.characteristics:
				try:
					read_val = await self.bleak_client.read_gatt_char(v2.uuid)
					read_val = ':'.join(['%02x'%b for b in read_val]).upper()
				except Exception as e:
					read_val = 'cannot read: ' + str(e)
				self.logger.debug('\tcharacteristic %s = [%s]', v2, read_val)
				for v3 in v2.descriptors:
					self.logger.debug('\t\tdescriptor %s handle %d', v3, v3.handle)
		for k, v in vars(Registers).items():
			if not isinstance(v, Register):
				continue
			try:
				read_val = await self.bleak_client.read_gatt_char(v.address)
				read_val = ':'.join(['%02x'%b for b in read_val]).upper()
			except Exception as e:
				read_val = 'cannot read: ' + str(e)
	
			self.logger.debug('REGISTER %s -> %s = [%s]', v.address, v.name, read_val)


	@synchronized(SYNC_LOCK)
	async def disconnect(self):
		'''
		Disconnects from device.
		Automatically cancels pending notifications.
		'''
		if not self.connected:
			self.logger.warning('client is not connected')
			return
		
		await self._shutdown_notifications()

		await self._disconnect()
	
	
	@synchronized(SYNC_LOCK)
	async def shutdown(self):
		'''
		Shutdown the device and disconnects.
		Automatically cancels pending notifications.
		'''
		await self.assert_authorized()

		await self._shutdown_notifications()

		self.logger.info('sending device shutdown signal')
		await self.write(Registers.SHUTDOWN, True)
		
		await self._disconnect()


	async def _disconnect(self):
		self.logger.info('disconnecting from device')

		if self.bleak_client:
			await self.bleak_client.disconnect()

		self.connected = False
		self.bleak_client = None
		self.connected_address = None

		self.logger.info('disconnected from device')


	async def _shutdown_notifications(self):
		'''
		Attempts cancellation of notification subscriptions.
		Fails silently in case of cancellation error.
		'''
		for register_key, callbacks in self.user_handlers.items():
			self.logger.debug('unregistering all callbacks for register %s', register_key)
			for callback_handler in callbacks:
				try:
					await self.unregister(callback_handler)
				except Exception as e2:
					self.logger.exception('error closing callback handler for register %s: %s', register_key, e2)


	def is_connected(self):
		'''
		Checks wether connection to the device is active
		'''
		return self.connected


	def assert_connected(self):
		'''
		Shortcut method to raise error if the device is not connected
		'''
		if not self.connected:
			raise ValueError('Client is not connected')
		

	async def is_authorized(self):
		'''
		Checks wether the user has authorized the connection by pressing the central button
		'''
		return self.is_connected() and await self.get_key_state(silent=True) 
	

	async def assert_authorized(self):
		'''
		Shortcut method to raise error if the connection has not been authorized by pressing the central button
		'''
		if self.key_state_check:
			if not await self.is_authorized():
				raise ValueError('Client is not authorized (KEY_STATE = 0). Press the central button to proceed')
		else:
			self.logger.debug('skipping key_state check because it is disabled')


	def enable_key_state_check(self):
		'''
		Enable check of KEY_STATE before attempting commands.
		'''
		self.key_state_check = True


	def disable_key_state_check(self):
		'''
		Disable check of KEY_STATE before attempting commands.
		'''
		self.key_state_check = False


	async def read(self, register, silent=False):
		'''
		Reads from the specified register.
		Automatically converts the data from device format to user format using the register configuration.
		The silent parameters allow to skip logging for recurrent background reads.
		'''
		self.assert_connected()
		
		if not register:
			raise ValueError('Register is required')

		self.logger_io.debug('READING %s %s', register.address, register.name)
		read_direct = await self.bleak_client.read_gatt_char(register.address)
		converted = register.from_device(read_direct, silent=silent)
		self.logger_io.info('READ %s %s >> %s', register.address, register.name, converted)
		return converted


	async def write(self, register, value=None):
		'''
		Writes to the specified register.
		Automatically converts the data from user format to device format using the register configuration.
		'''
		self.assert_connected()
		
		if not register:
			raise ValueError('Register is required')

		converted_value = register.to_device(value)

		self.logger_io.info('WRITE %s %s << %s', register.address, register.name, converted_value)
		
		await self.bleak_client.write_gatt_char(register.address, bytearray(converted_value))


	async def get_manufacturer_name(self):
		'''
		Reads the manufacturer name.
		Returns an ASCII UTF-8 string
		'''
		self.assert_connected()
		
		return await self.read(Registers.MANUFACTURER_NAME)


	async def get_model_number(self):
		'''
		Reads the model number.
		Returns an ASCII UTF-8 string
		'''
		self.assert_connected()
		
		return await self.read(Registers.MODEL_NUMBER)


	async def get_hardware_revision(self):
		'''
		Reads the device hardware revision.
		Returns an ASCII UTF-8 string
		'''
		self.assert_connected()
		
		return await self.read(Registers.HARDWARE_REVISION)


	async def get_firmware_revision(self):
		'''
		Reads the device firmware revision.
		Returns an ASCII UTF-8 string
		'''
		self.assert_connected()
		
		return await self.read(Registers.FIRMWARE_REVISION)


	async def get_software_revision(self):
		'''
		Reads the device software revision.
		Returns an ASCII UTF-8 string
		'''
		self.assert_connected()
		
		return await self.read(Registers.SOFTWARE_REVISION)


	async def get_mac_address(self):
		'''
		Reads the device mac address.
		Returns an ASCII UTF-8 string in the format AA:BB:CC:DD:EE:FF
		'''
		self.assert_connected()
		
		return await self.read(Registers.MAC_ADDRESS)


	async def get_serial_number(self):
		'''
		Reads the device serial number.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		self.assert_connected()
		
		return await self.read(Registers.SERIAL_NUMBER)


	async def get_chip_id(self):
		'''
		Reads the device CHIP ID.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		self.assert_connected()
		
		return await self.read(Registers.CHIP_ID)


	async def get_device_name(self):
		'''
		Reads the device name.
		Returns an ASCII UTF-8 string
		'''
		self.assert_connected()
		
		# return await self.read(Registers.GENERIC_ACCESS_DEVICE_NAME)
		# macOS compatibility patch
		return await self.read(Registers.MODEL_NUMBER)


	async def get_system_id(self):
		'''
		Reads the system id.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		self.assert_connected()
		
		return await self.read(Registers.DEVICE_INFORMATION_SYSTEM_ID)


	async def get_pnp_id(self):
		'''
		Reads the PNP id.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		self.assert_connected()
		
		return await self.read(Registers.DEVICE_INFORMATION_PNP_ID)


	async def get_ieee_11073_20601(self):
		'''
		Reads the IEEE 11073 20601 regulation mandatory specification.
		Returns an ASCII UTF-8 string in the format XX:......:YY
		'''
		self.assert_connected()
		
		return await self.read(Registers.DEVICE_INFORMATION_IEEE11073)


	async def get_battery_level(self):
		'''
		Reads the battery level.
		Returns an integer from 0 to 100
		'''
		self.assert_connected()
		
		return await self.read(Registers.BATTERY_LEVEL)


	async def get_use_count(self):
		'''
		Reads the use count.
		Returns an integer.
		'''
		await self.assert_authorized()
		return await self.read(Registers.USER_RECORD)


	async def reset_use_count(self):
		'''
		Resets device use count to zero.
		'''
		await self.assert_authorized()
		self.logger.debug('resetting use count')
		return await self.write(Registers.USER_RECORD_RESET)


	async def get_buttons_status(self):
		'''
		Reads the button status. 
		Returns an int value from the enum Buttons:
			Buttons.NONE_PRESSED = 0x03
			Buttons.CENTRAL = 0x00
			Buttons.PLUS = 0x01
			Buttons.MINUS = 0x02
		'''
		await self.assert_authorized()
		return await self.read(Registers.BUTTON)


	async def get_temperature_and_pressure(self):
		'''
		Reads the device internal temperature and pressure.
		Returns a touple (temperature, pressure).
		Temperature is a floating point number with 2 decimals in Celsius degrees.
		Pressure is a floating point number with 2 decimals in mbar.
		Eg. (24.25, 987.23)
		'''
		await self.assert_authorized()
		return await self.read(Registers.PRESSURE_TEMPERATURE)


	async def get_temperature(self):
		'''
		Reads the device internal temperature.
		Returns a floating point number with 2 decimals in Celsius degrees.
		Eg. (24.25, 987.23)
		'''
		await self.assert_authorized()
		return (await self.read(Registers.PRESSURE_TEMPERATURE))[0]


	async def get_pressure(self):
		'''
		Reads the device internal pressure.
		Returns a floating point number with 2 decimals in mbar.
		Eg. 987.23
		'''
		await self.assert_authorized()
		return (await self.read(Registers.PRESSURE_TEMPERATURE))[1]


	async def get_accelerometer(self):
		'''
		Reads the device accelerometer data.
		Returns a touple (x, y, z) where x, y and z are integers.
		Eg. (10, 200, 1000)
		'''
		await self.assert_authorized()
		return await self.read(Registers.ACCELEROMETER)


	async def get_accelerometer_x(self):
		'''
		Reads the device accelerometer data for X axis.
		Returns an integer.
		'''
		await self.assert_authorized()
		return (await self.read(Registers.ACCELEROMETER))[0]


	async def get_accelerometer_y(self):
		'''
		Reads the device accelerometer data for Y axis.
		Returns an integer.
		'''
		await self.assert_authorized()
		return (await self.read(Registers.ACCELEROMETER))[1]


	async def get_accelerometer_z(self):
		'''
		Reads the device accelerometer data for Z axis.
		Returns an integer.
		'''
		await self.assert_authorized()
		return (await self.read(Registers.ACCELEROMETER))[2]


	async def get_depth(self):
		'''
		Reads the insertion depth level.
		Returns an integer from 0 to 8.
		'''
		await self.assert_authorized()
		return await self.read(Registers.LENGTH)


	async def get_rotation_speed(self):
		'''
		Reads the rotation speed from Hall sensors.
		Returns an integer representing the rotations per second.
		'''
		await self.assert_authorized()
		return await self.read(Registers.HALL)


	async def get_wake_up(self):
		'''
		Reads the quick wake-up status.
		Returns a boolean.
		'''
		await self.assert_authorized()
		return await self.read(Registers.WAKE_UP)


	async def enable_wake_up(self):
		'''
		Enables quick wake-up.
		'''
		await self.assert_authorized()
		value = True
		self.logger.info('setting wake-up to %s', value)
		return await self.write(Registers.WAKE_UP, value)


	async def disable_wake_up(self):
		'''
		Disables quick wake-up.
		'''
		await self.assert_authorized()
		value = False
		self.logger.info('setting wake-up to %s', value)
		return await self.write(Registers.WAKE_UP, value)


	async def get_vibration_setting(self):
		'''
		Reads the device auto vibration settings
		Returns a touple of 8 integers ranging from 0 to 100.
		'''
		await self.assert_authorized()
		return await self.read(Registers.VIBRATOR_SETTING)


	async def set_vibration_setting(self, value):
		'''
		Sets the auto vibration settings.
		Takes a collection of 8 integers ranging from 0 to 100.
		'''
		if not value:
			raise ValueError('Value is required')
		if len(value) != 8:
			raise ValueError('A collection of 8 elements is required')
		for v in value:
			if v < 0 or v > 100:
				raise ValueError('Value should be between 0 and 100')
		
		await self.assert_authorized()
		self.logger.info('setting vibration-setting to %s', value)
		return await self.write(Registers.VIBRATOR_SETTING, value)


	async def get_cruise_control(self):
		'''
		Reads the cruise control status.
		Returns a boolean.
		'''
		await self.assert_authorized()
		return await self.read(Registers.MOTOR_WORK_ON_TOUCH)


	async def disable_cruise_control(self):
		'''
		Disables cruise control.
		'''
		await self.assert_authorized()

		value = CruiseControlStatus.DISABLED
		self.logger.info('setting cruise-control to %s', value)
		return await self.write(Registers.MOTOR_WORK_ON_TOUCH, value)


	async def enable_cruise_control(self, reset=False):
		'''
		Enables cruise control.
		If reset=True is passed, resets motors speed to default value.
		'''
		await self.assert_authorized()

		value = CruiseControlStatus.ENABLE_AND_RESET if reset else CruiseControlStatus.ENABLE
		self.logger.info('setting cruise-control to %s', value)
		return await self.write(Registers.MOTOR_WORK_ON_TOUCH, value)


	async def get_key_state(self, silent=False):
		'''
		Reads the key state status.
		Returns a boolean.
		'''
		self.assert_connected()
		if not silent:
			self.logger.debug('checking key state')
		return await self.read(Registers.KEY_STATE, silent=silent)


	async def stop_motors(self):
		'''
		Stops both motors of the device.
		'''
		await self.assert_authorized()
		
		self.logger.info('sending motors stop signal')
		await self.write(Registers.MOTOR_STOP, True)


	async def verify_accelerometer(self):
		'''
		Puts the device into VERIFY ACCELEROMETER mode.
		'''
		await self.assert_authorized()
		self.logger.info('enterying accelerometer verification mode')
		return await self.write(Registers.VERIFY_ACCELEROMETER)
	
	
	async def get_motors_speed(self):
		'''
		Reads the two motors speed.
		Returns a touple (X, Y) with X being the main motor speed ranging from 0 to 100
		and Y being the vibrator motor speed ranging from 0 to 100
		'''
		await self.assert_authorized()
		return await self.read(Registers.MOTOR_SPEED)


	async def get_main_motor_speed(self):
		'''
		Reads the main motor speed.
		Returns an integer ranging from 0 to 100
		'''
		await self.assert_authorized()
		return (await self.read(Registers.MOTOR_SPEED))[0]


	async def get_vibration_speed(self):
		'''
		Reads the vibrator motor speed.
		Returns an integer ranging from 0 to 100
		'''
		await self.assert_authorized()
		return (await self.read(Registers.MOTOR_SPEED))[1]


	async def set_motors_speed(self, value):
		'''
		Set the two motors speed.
		Takes a collection of two elements (X, Y) 
		with X being the main motor speed ranging from 0 to 100
		and Y being the vibrator motor speed ranging from 0 to 100
		'''
		if not value:
			raise ValueError('Value is required')
		if len(value) != 2:
			raise ValueError('A collection of 2 elements is required')
		for v in value:
			if v < 0 or v > 100:
				raise ValueError('Value should be between 0 and 100')
		
		await self.assert_authorized()
		self.logger.info('setting motor speed to %d and vibration speed to %d', value[0], value[1])
		return await self.write(Registers.MOTOR_SPEED, value)


	async def set_main_motor_speed(self, value):
		'''
		Sets the main motor speed.
		Takes an integer ranging from 0 to 100
		'''
		if value is None:
			raise ValueError('Value is required')
		if value < 0 or value > 100:
			raise ValueError('Value should be between 0 and 100')
		
		await self.assert_authorized()
		self.logger.info('setting main motor speed to %d', value)
		current_value = await self.get_vibration_speed();
		self.logger.debug('will mantain current vibration speed of %d', current_value)
		return await self.write(Registers.MOTOR_SPEED, [value, current_value])


	async def set_vibration_speed(self, value):
		'''
		Sets the vibrator motor speed.
		Takes an integer ranging from 0 to 100
		'''
		if value is None:
			raise ValueError('Value is required')
		if value < 0 or value > 100:
			raise ValueError('Value should be between 0 and 100')
		
		await self.assert_authorized()
		self.logger.info('setting vibration speed to %d', value)
		current_value = await self.get_main_motor_speed();
		self.logger.debug('will mantain current motor speed of %d', current_value)
		return await self.write(Registers.MOTOR_SPEED, [current_value, value])


	async def ping(self):
		'''
		Reads from a register to test device communication
		'''
		self.assert_connected()

		self.logger.debug('sending ping to device')
		await self.read(Registers.MODEL_NUMBER, silent=True)
		self.logger.debug('device responded to ping')
	
	
	async def notify_key_state(self, user_callback, distinct_until_changed=True):
		'''
		Subscribes to key state change notification.
		Requires a callback that accepts a boolean representing the key state value.
		Returns an handler.
		
		If distinct_until_changed=True, notifications will be emitted only on value changes.
		If distinct_until_changed=False, multiple spurious notifications will be emitted.
		
		To stop receiving the notifications, call .unregister() on the returned handler.
		Alternatively you can call .unregister(handler) on this client.
		'''
		self.assert_connected()
		
		self.logger.debug('registering callback for key state')
		cb = await self._create_callback_handler(Registers.KEY_STATE, user_callback, distinct_until_changed=distinct_until_changed)
		return cb
	
	
	async def notify_buttons(self, user_callback, distinct_until_changed=True):
		'''
		Subscribes to pressed buttons change notification.
		Requires a callback that accepts an integer representing the pressed buttons status from the Buttons enum:
			Buttons.NONE_PRESSED = 0x03
			Buttons.CENTRAL = 0x00
			Buttons.PLUS = 0x01
			Buttons.MINUS = 0x02
			
		Returns an handler.
		
		If distinct_until_changed=True, notifications will be emitted only on value changes.
		If distinct_until_changed=False, multiple spurious notifications will be emitted.
		
		To stop receiving the notifications, call .unregister() on the returned handler.
		Alternatively you can call .unregister(handler) on this client.
		'''
		self.assert_connected()
		
		self.logger.debug('registering callback for buttons')
		cb = await self._create_callback_handler(Registers.BUTTON, user_callback, distinct_until_changed=distinct_until_changed)
		return cb
	
	
	async def notify_rotation_speed(self, user_callback, distinct_until_changed=True):
		'''
		Subscribes to rotation speed change notification.
		Requires a callback that accepts an integer representing the rotation speed in rotations per second.
		Returns an handler.
		
		If distinct_until_changed=True, notifications will be emitted only on value changes.
		If distinct_until_changed=False, multiple spurious notifications will be emitted.
		
		To stop receiving the notifications, call .unregister() on the returned handler.
		Alternatively you can call .unregister(handler) on this client.
		'''
		self.assert_connected()
		
		self.logger.debug('registering callback for rotation speed')
		cb = await self._create_callback_handler(Registers.HALL, user_callback, distinct_until_changed=distinct_until_changed)
		return cb
	
	
	async def notify_depth(self, user_callback, distinct_until_changed=True):
		'''
		Subscribes to depth insertion change notification.
		Requires a callback that accepts an integer representing depth insertion from 0 to 8.
		Returns an handler.
		
		If distinct_until_changed=True, notifications will be emitted only on value changes.
		If distinct_until_changed=False, multiple spurious notifications will be emitted.
		
		To stop receiving the notifications, call .unregister() on the returned handler.
		Alternatively you can call .unregister(handler) on this client.
		'''
		self.assert_connected()
		
		self.logger.debug('registering callback for depth')
		cb = await self._create_callback_handler(Registers.LENGTH, user_callback, distinct_until_changed=distinct_until_changed)
		return cb
	
	
	async def notify_accelerometer(self, user_callback, distinct_until_changed=True):
		'''
		Subscribes to accelerometer data change notification.
		Requires a callback that accepts a touple of 3 integers (see get_accelerometers return value).
		Returns an handler.
		
		If distinct_until_changed=True, notifications will be emitted only on value changes.
		If distinct_until_changed=False, multiple spurious notifications will be emitted.
		
		To stop receiving the notifications, call .unregister() on the returned handler.
		Alternatively you can call .unregister(handler) on this client.
		'''
		self.assert_connected()
		
		self.logger.debug('registering callback for accelerometer')
		cb = await self._create_callback_handler(Registers.ACCELEROMETER, user_callback, distinct_until_changed=distinct_until_changed)
		return cb
	
	
	async def notify_temperature_and_pressure(self, user_callback, distinct_until_changed=True):
		'''
		Subscribes to temperature and pressure data change notification.
		Requires a callback that accepts a touple of 2 integers (see get_temperature_and_pressure return value).
		Returns an handler.
		
		If distinct_until_changed=True, notifications will be emitted only on value changes.
		If distinct_until_changed=False, multiple spurious notifications will be emitted.
		
		To stop receiving the notifications, call .unregister() on the returned handler.
		Alternatively you can call .unregister(handler) on this client.
		'''
		self.assert_connected()
		
		self.logger.debug('registering callback for pressure and temperature')
		cb = await self._create_callback_handler(Registers.PRESSURE_TEMPERATURE, user_callback, distinct_until_changed=distinct_until_changed)
		return cb
	
	
	@synchronized(SYNC_LOCK)
	async def _create_callback_handler(self, register, user_callback, distinct_until_changed=False):
		'''
		Internal use only.
		Creates a callback handler and registers BLE-level notification channel if needed.
		'''
		self.logger_callback.debug('creating callback handler for register %s %s', register.address, register.name)
		
		# Create the handler
		o = CallbackHandler(self, register, user_callback)

		if not register.address in self.user_handlers:
			self.user_handlers[register.address] = []
		registered_be_handlers = self.user_handlers[register.address]
		registered_be_handlers.append(o)
		
		self.logger_callback.debug('there are now %d callbacks listening for register %s %s', len(registered_be_handlers), register.address, register.name)

		# If this is the first callback for the characteristic, enable BLE notification channel
		if not register.address in self.bleak_handlers:
			self.logger_callback.debug('backend callback for register %s %s is not enabled (first listener), enabling now', register.address, register.name)
			def callback(sender, data):
				self._dispatch_notification(register, sender, data, distinct_until_changed=distinct_until_changed)

			self.bleak_handlers[register.address] = register.address
			await self.bleak_client.start_notify(register.address, callback)
			self.logger_callback.debug('registered backend callback for register %s %s', register.address, register.name)
			
		# Activate the handler before returning
		o.activate()
		
		self.logger_callback.info('activated user callback handler for register %s %s', register.address, register.name)
		return o


	def _dispatch_notification(self, register, sender, raw_data, distinct_until_changed=False):
		'''
		Internal use only.
		Dispatch notification to user callbacks.
		'''
		if distinct_until_changed:
			if not self._has_changed(register, raw_data):
				return

		self.logger_callback.debug('received notification of data change for register %s %s', register.address, register.name)
		converted = register.from_device(raw_data)
		self.logger_io.debug('NOTIFICATION %s %s >> %s', register.address, register.name, converted)
			
		if register.address in self.user_handlers and len(self.user_handlers[register.address]) > 0:
			for user_handler in self.user_handlers[register.address]:
				if user_handler.is_active():
					self.logger_callback.debug('dispatching notification of data change to user callback')
					user_handler.dispatch(converted)
				else:
					self.logger_callback.debug('skipping notification of data change to user callback because it is not active')
		else:
			self.logger_callback.warning('received notification from register %s but no user callbacks', register.address)


	def _has_changed(self, register, raw_data):
		'''
		Internal use only.
		Checks if register data has changed since last notification.
		Note: this method is long and branched because it is called a lot of times and needs to run faster.
		'''
		changed = False
		# self.logger_callback.debug('checking if data changed')
		
		if not register.address in self.callback_data_history:
			changed = True
		else:
	
			data_prv = self.callback_data_history[register.address]
			data_now_none = raw_data is None or len(raw_data) < 1
			data_prv_none = data_prv is None or len(data_prv) < 1
			
			if data_now_none and data_prv_none:
				return False
			elif not data_now_none and data_prv_none:
				changed = True
			elif data_now_none and not data_prv_none:
				changed = True
			elif len(raw_data) != len(data_prv):
				changed = True
			else:
				len_now = len(raw_data)
				i = 0
				while i < len_now:
					if raw_data[i] != data_prv[i]:
						changed = True
						break
					i = i + 1
			
		if changed:
			self.callback_data_history[register.address] = raw_data

		# self.logger_callback.debug('checking data changed resulted %s', changed)
		return changed


	@synchronized(SYNC_LOCK)
	async def unregister(self, callback_handler):
		'''
		Unregisters a callback handler and unregisters BLE-level notification channel if no longer needed.
		'''
		await self._unregister(callback_handler)
		

	async def _unregister(self, callback_handler):
		'''
		Internal use only.
		Unregisters a callback handler and unregisters BLE-level notification channel if no longer needed.
		'''
		if not callback_handler:
			raise ValueError('Callback handler to unregister is required')
		
		# Deactivate the handler
		callback_handler.deactivate()
		
		register = callback_handler.register
		self.logger_callback.debug('unregistering callback handler for register %s %s', register.address, register.name)
		callback_handler.deactivate()
		
		registered_be_handlers = self.user_handlers[register.address]
		registered_be_handlers.remove(callback_handler)
		
		self.logger_callback.debug('there are now %d callbacks listening for register %s %s', len(registered_be_handlers), register.address, register.name)
	
		# If this was the last callback for the characteristic, disable BLE notification channel
		if len(registered_be_handlers) < 1 and register.address in self.bleak_handlers:
			self.logger_callback.debug('backend callback for register %s %s is no longer needed, unregistering', register.address, register.name)
			
			self.bleak_handlers[register.address] = register.address
			await self.bleak_client.stop_notify(register.address)
			del self.bleak_handlers[register.address]
			self.logger_callback.debug('unregistered backend callback for register %s %s', register.address, register.name)

		self.logger_callback.info('deactivated user callback handler for register %s %s', register.address, register.name)


import logging
from .definitions import *


class CallbackHandler:
	'''
	Basic class to handle notifications callbacks.
	An instance of this object is returned to caller when subscribing to an event.
	The caller can .unregister using either the client or the returned object.
	'''
	logger = logging.getLogger(Constants.LOGGER_CALLBACK_NAME)

	def __init__(self, original_client, register, user_callback):
		self.original_client = original_client
		self.register = register
		self.user_callback = user_callback

	def activate(self):
		'''
		Activates reception of notifications
		'''
		self.active = True
	
	def deactivate(self):
		'''
		Temporarily deactivates reception of notifications.
		Note that deactivating a callback handler does not unsubscribe at the BLE level.
		BLE-level notifications unsubscription is automatically managed when calling unregister().
		To remove permanently remove subscription use the unregister() method.
		'''
		self.active = False
	
	def is_active(self):
		'''
		Checks wether the callback should receive notification data.
		'''
		return self.active	
		
	async def unregister(self):
		'''
		Permanently deactivates the callback.
		If no other callbacks are still active on the register, unregisters from notifications at BLE-level.
		'''
		return await self.original_client.unregister(self)

	def dispatch(self, data):
		'''
		Dispatch notification data to user-provided callback.
		'''
		if not self.is_active():
			self.logger.warning('dispatch called but handler is not active, skipping')
			return
		if not self.user_callback:
			self.logger.warning('dispatch called but no callback provided, skipping')
			return
		self.logger.debug('dispatching to final user handler started')
		
		try:
			self.user_callback(data)
		except Exception as e:
			self.logger.error('Error dispatching notification to user handler: %s', e)

		self.logger.debug('dispatching to final user handler completed')


import asyncio
import time
import logging
import socket
import json
import re

from .definitions import *
from .async_client import *
from asyncio.futures import CancelledError


class SocketConnection:
	'''
	Holds information for connected clients.
	'''

	id = None
	name = None
	authorized = False
	locking = False

	def __init__(self, id):
		self.id = id
		
	def is_authorized(self):
		return self.authorized
	
	def authorize(self):
		self.authorized = True


class SocketServer():
	'''
	Asynchronous LELO F1 SDK SOCKET server.
	Configurable with standard logging.
	'''
	logger = logging.getLogger(Constants.LOGGER_SOCKET_SERVER_NAME)
	logger_io = logging.getLogger(Constants.LOGGER_IO_NAME)
	
	SOCKET_ENCODING = 'utf8'

	def __init__(self, client_provider, ip_address, port):
		'''
		Instantiate the socket server.
		Takes a client or a client provider as only parameter.
		'''
		self.logger.debug('instantiating LELO F1 SDK SOCKET server')
		self.stopped = False
		self.stop_requested = False
		
		if isinstance(client_provider, AsyncClient):
			self.client_provider = None
			self.client = client_provider
		else:
			self.client_provider = client_provider
			self.client = None
		
		self.ip_address = ip_address
		self.port = port
		self.clients = {}
		self.blocked_clients = []
		
		self.authorization_handler = None
		self.command_handler = None

		self.callbacks = dict()


	def set_authorization_handler(self, handler):
		'''
		Set the async handler to be used to authorize client connections.
		'''
		self.authorization_handler = handler


	def set_command_handler(self, handler):
		'''
		Set the async handler to be used to execute commands.
		'''
		self.command_handler = handler

	
	def on_after_command(self, callback):
		'''
		Register a function to be called after every succesfull command 
		'''
		self._register_callback('AFTER_COMMAND', callback)

	
	def on_client_connected(self, callback):
		'''
		Register a function to be called when a client disconnects
		'''
		self._register_callback('CLIENT_CONNECT', callback)

	
	def on_client_disconnect(self, callback):
		'''
		Register a function to be called when a client disconnects
		'''
		self._register_callback('CLIENT_DISCONNECT', callback)


	async def _dispatch_callback(self, key, *args):
		'''
		Internal use only.
		'''
		if not key in self.callbacks:
			return
		for cb in self.callbacks[key]:
			try:
				await cb(*args)
			except Exception as e:
				self.logger.exception('error in user callback: %s', e)


	def _register_callback(self, key, cb):
		'''
		Internal use only.
		'''
		if not key in self.callbacks:
			self.callbacks[key] = []
		self.callbacks[key].append(cb)
		

	def get_active_clients(self):
		'''
		Returns a list of SocketConnection objects representing the active clients.
		'''
		return self.clients.values()


	def get_client(self):
		'''
		Get the device client from the fixed one passed at construction or from the provider.
		'''
		if self.client_provider:
			return self.client_provider()
		elif self.client:
			return self.client
		else:
			return None


	def stop(self):
		self.logger.debug('scheduling socket server stop request')
		if not self.stop_requested:
			self.stop_requested = True
			self.stopped = True


	async def run_server(self, loop):
		'''
		Socket server coroutine.
		Runs inside asyncio event loop passed as parameter.
		'''
		try:
			self.logger.info('running socket server')
			self.loop = loop
		
			self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.logger.info('opening socket on %s:%d', self.ip_address, self.port)
			
			self.socket.bind((self.ip_address, self.port))
			self.socket.listen(1) 
			self.socket.setblocking(False)
			
			self.logger.info('socket server waiting for connection')
			while not self.stopped:
				try:
					client, _ = await asyncio.wait_for(
		                self._accept_connection(),
		                timeout=3,
		                loop=self.loop,
		            )
				except asyncio.futures.TimeoutError:
					if self.stopped:
						break
					continue

				except CancelledError as sock_accept_exception:
					self.logger.warning('socket accept operation cancelled: %s', sock_accept_exception)
					break
				except Exception as sock_accept_exception:
					self.logger.exception('socket GENERIC EXCEPTION: %s', sock_accept_exception)
					break
				
				self.logger.info('socket server received a connection from ' + str(client.getpeername()))
				self.loop.create_task(self.handle_client(client))

		except Exception as ext:
			self.logger.exception('error running socket server: %s', ext)

		if False and self.socket:
			self.logger.debug('closing socket following close request')
			self.socket.close()
			self.logger.debug('closed socket following close request')

		self.logger.info('socket server stopped')


	async def _accept_connection(self):
		return await self.loop.sock_accept(self.socket)


	async def _receive_connection(self, client, limit):
		return await self.loop.sock_recv(client, limit)


	async def handle_client(self, client):
		'''
		Handles commands from a client.
		Automatically translates data from/to JSON <-> dictionary.
		'''
		client_info = self.get_or_create_client_info(client)
		self.logger.info( (client_info.name or client_info.id) + ' CONNECTED')
		
		await self._dispatch_callback('CLIENT_CONNECT', client_info)
		
		try:
			self.logger.info('socket server serving client %s', client.getpeername())
			
			request = None
			
			while not self.stopped:
				try:
					request_raw = await asyncio.wait_for(
		                self._receive_connection(client, 1023),
		                timeout=3,
		                loop=self.loop,
		            )
				except asyncio.futures.TimeoutError:
					if self.stopped:
						break
					continue

				if request_raw is None or len(request_raw) == 0:
					self.logger.info('received empty packet, disconnecting')
					break
				request = request_raw.decode(self.SOCKET_ENCODING)
				try:
					payload = json.loads(request)
					
					self.logger_io.debug('RECEIVED %s', request)
					
					response = json.dumps(await self.handle_command(client_info, payload))
					
				except Exception as e:
					self.logger.exception('error handling client command: %s', e)
					response = json.dumps(self.response_error(e))

				self.logger_io.debug('SEND %s', response)
				
				await self.loop.sock_sendall(client, (response + '\n').encode(self.SOCKET_ENCODING))
				
				if not client_info.id in self.clients:
					self.logger.info('command requested disconnection')
					break

			self.logger.info('socket server closing connection')
			client.close()
			
		except CancelledError as sock_accept_exception:
			self.logger.warning('socket read operation cancelled: %s', sock_accept_exception)
		except Exception as ext:
			self.logger.exception('error handling client: %s', ext)
		finally:
			if client_info.id in self.clients:
				del self.clients[client_info.id]
			self.logger.info( (client_info.name or client_info.id) + ' DISCONNECTED')
		
		await self._dispatch_callback('CLIENT_DISCONNECT', client_info)


	def get_or_create_client_info(self, client):
		client_id = client.getpeername()[0] + ':' + str(client.getpeername()[1])
		if client_id in self.clients:
			self.logger.debug('client %s is known already', client_id)
			return self.clients[client_id]
		else:
			self.logger.debug('client %s is a new client, registering', client_id)
			client_info = SocketConnection(client_id)
			self.clients[client_id] = client_info
			return client_info


	async def handle_command(self, client_info, payload):
		'''
		Handle a single command received from socket.
		client_id is the client identifier in 'IP:PORT' format.
		payload is a dictionary obtained from json deserialization.
		'''
		try:
			if not payload or not 'command' in payload:
				raise ValueError('Malformed command: ' + str(payload))

			command = payload['command'].lower() if 'command' in payload else None
			arguments = payload['arguments'] if 'arguments' in payload else []

			self.logger.info('received command: ' + (command) + (' from ' + client_info.name if client_info.name else ''))
			
			data = await self.execute_command(client_info, command, arguments)
			
			await self._dispatch_callback('AFTER_COMMAND', client_info, command, arguments, data)

			return self.response_ok(data)

		except Exception as e:
			self.logger.exception('error handling command: %s', e)
			return self.response_error(e)


	async def execute_command(self, client_info, command, arguments):
		'''
		Executes the command on local first and on device if not a local command.
		'''
		command = re.sub(r'[^a-z0-9_]', '', command.lower())

		self.logger.debug('attempting to run command %s as a managed command', command)
		handled, data = await self.execute_command_managed(client_info, command, arguments)
		
		if not handled and self.command_handler:
			self.logger.debug('attempting to run command %s from registered command handler', command)
			handled, data = await self.command_handler(client_info, command, arguments)
		
		if not handled:
			self.logger.debug('attempting to run command %s as a device command', command)
			data = await self.execute_command_device(client_info, command, arguments)
		
		return data


	async def execute_command_managed(self, client_info, command, arguments):
		'''
		Executes a managed command.
		'''
		if hasattr(self, 'command_' + command):
			return (True, await getattr(self, 'command_' + command)(client_info, arguments))
		else:
			self.logger.debug('command %s is not a managed command', command)
			return (False, None)


	async def execute_command_device(self, client_info, command, arguments):
		'''
		Executes a command directly on the device backend.
		'''
		self.assert_authorized(client_info)
		
		if command.startswith('_'):
			raise ValueError('Forbidden command ' + command)

		if command.startswith('notif'):
			raise ValueError('Forbidden notification command ' + command)
		
		handler = getattr(self.get_client(), command)
		if not handler:
			raise ValueError('Unknown command ' + command)

		data = await handler(*arguments)
		return data
	
	
	def assert_authorized(self, client_info):
		'''
		Shortcut method to check that the client has been authorized.
		'''
		if not client_info or not client_info.is_authorized():
			raise ValueError('client is not authorized')


	def response_ok(self, data=None):
		out = dict()
		out['status'] = 'OK'
		if data is not None:
			out['data'] = data
		return out


	def response_error(self, e):
		out = dict()
		out['status'] = 'ERROR'
		out['message'] = str(e)
		return out


	async def command_ping(self, client_info, arguments):
		'''
		Checks that the connection works.
		'''
		return 'pong'


	async def command_name(self, client_info, arguments):
		'''
		Associates a name to the client.
		'''
		client_info.name = str(arguments[0])


	async def command_quit(self, client_info, arguments):
		'''
		Closes the connection.
		'''
		del self.clients[client_info.id]


	async def command_authorized(self, client_info, arguments):
		'''
		Checks wether client has been authorized.
		'''
		if client_info.id in self.blocked_clients:
			return 'BLOCKED'
		else:
			return client_info.authorized


	async def command_connected(self, client_info, arguments):
		'''
		Checks wether client device is connected.
		'''
		return self.get_client().is_connected()
	
	
	async def command_authorize(self, client_info, arguments):
		'''
		Requests user authorization if needed
		'''
		if client_info.authorized:
			return True
		
		if client_info.id in self.blocked_clients:
			return False
		
		if not self.authorization_handler:
			raise ValueError('No authorization handler registered')

		result = await self.authorization_handler(client_info, arguments)
		
		identifier = client_info.name or client_info.id
		
		if result:
			self.logger.info('granted control to ' + identifier)
			client_info.authorized = True
		else:
			self.logger.warn('blocked control from ' + identifier)
			self.blocked_clients.append(client_info.id)



from lelof1py.definitions import *
from lelof1py.register import *
from lelof1py.registers import *
from lelof1py.async_client import *
from lelof1py.socket_server import *
from lelof1py.socket_client import *
from lelof1py.callback_handler import *
from lelof1py.helpers import *
from lelof1py.fs_utils import *



from distutils.core import setup


setup(
	name = 'lelof1py',
	packages = ['lelof1py'],
	version = '0.10-beta',
	license = 'MIT',
	description = 'Python client for LELO F1 SDK device',
    long_description = 'See https://github.com/fabiofenoglio/lelo-f1-python-sdk',
	author = 'Fabio Fenoglio',
	author_email = 'development@fabiofenoglio.it',
	url = 'https://github.com/fabiofenoglio/lelo-f1-python-sdk',
	download_url = 'https://github.com/fabiofenoglio/lelo-f1-python-sdk/archive/v0.10-beta.tar.gz',
	keywords = ['LELO', 'BLUETOOTH', 'REMOTE'],
	install_requires = [
		'asyncio',
		'bleak',
        'appdirs'
	],
	classifiers = [
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'Topic :: Software Development :: Build Tools',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
	],
)
